var character;
var canvas = document.getElementById("cvGame");
var ctx = canvas.getContext("2d");
var game = {
    states: Object.freeze({"start": 0, "running":1, "over": 2,"end" : 3}),
    speed: 50,
    state: null,
    tmr: null,
    gravity: 0.75,
    jump: -14,
    shiftDown: false,
    leftDown: false,
    rightDown: false,
    ctrlDown: false
};
var sprites = {
    platforms: [],
    hazards: [],
    objects: [],
    coins: [],
    stops: []
};
var Sprig = function(){
    this.x = 0;
    this.y = 0;
    this.moveX = 0;
    this.moveY = 0;
    this.visible = false;
};
var Sprite = function(imgSrc){
    Sprig.call(this);
    this.img = new Image();
    this.img.src = imgSrc;
    this.continuous = false;
};
Sprite.prototype.getRect = function() {
    return {x: this.x, y: this.y, width: this.img.width, height: this.img.height};
};
var Anim = function(imgSrc, frames, frameWidth, repeat){
    this.img = new Image();
    this.img.src = imgSrc;
    this.frames = frames;
    this.frameWidth = frameWidth;
    this.frame = -1;
    this.default = 0;
    this.repeat = repeat;
    this.pingPong = false;
    this.pp = 1;
    this.reset = function() {this.frame = -1;};
};
Anim.prototype.animate = function(x, y) {
    if(game.state == game.states.running){
        if(this.repeat){
            if(this.pingPong){
                this.frame += this.pp;
                if(this.frame >= this.frames || this.frame < 0){
                    this.pp *= -1; // reverse
                    this.frame += this.pp * 2;
                }
            } else {
                this.frame++;
                this.frame %= this.frames;
            }
        } else {
            if(this.frame < this.frames -1){
                this.frame++;
            }
        }
    } else {
        this.frame = this.default; // stop animation
    }
    let clipx = this.frame * this.frameWidth;
    ctx.drawImage(this.img, clipx, 0, this.frameWidth, this.img.height, x, y, this.frameWidth, this.img.height);
};

Anim.prototype.getRect = function(x, y) {
    return {x: x, y: y, width: this.frameWidth, height: this.img.height};
};
var Actor = function(){
    Sprig.call(this);
    this.anims = [];
    this.current = 0;
    this.height = 0;
    this.resetCurrent = function(current) {
        this.current = current;
        this.anims[this.current].reset();
    };
};
Actor.prototype.manageActions = function() {};
Actor.prototype.collision = function() {};

Actor.prototype.getRect = function() {
    return this.anims[this.current].getRect(this.x, this.y);
};
Sprite.prototype.animate = function() {
    if(this.visible){
        ctx.drawImage(this.img, this.x, this.y);
    }
    if(game.state === game.states.running || game.state === game.states.over){
        this.x += this.moveX;
        this.y += this.moveY;
        if(this.continuous && this.x < (canvas.width - this.img.width)){
            this.x = 0;
        }
    }
};
Actor.prototype.animate = function() {
    if(this.visible){
        this.anims[this.current].animate(this.x, this.y);
    }
    if(game.state == game.states.running){
        this.x += this.moveX;
        this.y += this.moveY;
    }
    this.manageActions();
};
Actor.prototype.collisionDetect = function() {
    let mRect = this.getRect();
    let colls = [];
    for(let list in sprites){
        let spArray = sprites[list];
        for(let s = 0; s < spArray.length; s++){
            let sr = spArray[s].getRect();
            if(bump(mRect, sr)){
                colls.push({type: list, index: s});
            }
        }
    }
    if(colls.length){
        this.collision(colls);
    }
};

var background = {
    sprites: [],
    backgroundSpeed: -0.5,
    bgWidth: 1024,
    midbackSpeed: -2.25,
    midBackWidth: 1024,
    midBackHeight: 235,
    midBack: [],
};
function addBackgrounds(){
    let back = new Sprite("images/background.png");
    back.visible = true;
    back.x = 0;
    //back.moveX = background.backgroundSpeed;
    background.sprites.push(back);
    back = new Sprite("images/background.png");
    //back.x = background.bgWidth;
    back.moveX = background.backgroundSpeed;
    background.sprites.push(back);
}
function manageBackgrounds() {
    for(let b = 0; b < 2; b++) {
        background.sprites[b].animate();
    }
}
function initMidBackground() {
    let m = new Sprite("images/altmid.png");
    m.x = 0;
    m.y = canvas.height - (background.midBackHeight + 38);
    //m.moveX = background.midbackSpeed;
    m.visible = true;
    background.midBack.push(m);
}
function animateMiddleGround() {
    for(let m = 0; m < background.midBack.length; m++){
        background.midBack[m].animate();
    }
    if(background.midBack[0].x <= -background.midBackWidth){
        background.midBack[0].x = canvas.width;
    }
}
var platforms = {
    sprites: [],
    moveX: -5
};
function addPlatforms() {
    let w = 0;
    while(w <= canvas.width){
        addPlatform(w);
        w += 128;
    }
}
function managePlatforms() {
    for(let p = 0, q = platforms.sprites.length; p < q; p++){
        platforms.sprites[p].animate();
    }
    if(platforms.sprites[0].x <= -platforms.sprites[0].img.width){
        let p = platforms.sprites.shift(); // remove first array element
        p.x = platforms.sprites[platforms.sprites.length - 1].x + 128;
        platforms.sprites.push(p); // push it back to array end
    }
}
function addPlatform(xPos){
    let p = new Sprite("images/testPlatform.png");
    p.x = xPos;
    p.y = canvas.height - 128;
    p.visible = true;
    p.moveX = platforms.moveX;
    platforms.sprites.push(p);
}
function bump(rectA, rectB){
    return Math.max(rectA.x, rectB.x) < Math.min(rectA.x + rectA.width, rectB.x + rectB.width) &&
        Math.max(rectA.y, rectB.y) < Math.min(rectA.y + rectA.height, rectB.y + rectB.height);
}
function intersect(rectA, rectB){
    let rA = rectA.x + rectA.width, rB = rectB.x + rectB.width;//right x
    let bA = rectA.y + rectA.height, bB = rectB.y + rectB.height;//btm y
    let ix = Math.max(rectA.x, rectB.x), iy =  Math.max(rectA.y, rectB.y); //?? min
    let rect = {
        x: ix,
        y: iy,
        width: Math.min(rA, rB) - ix,
        height: Math.min(bA, bB) - iy
    };
    if(rect.x < (rect.x + rect.width) && rect.y < (rect.y + rect.height)){
        return rect;
    }
    return false;
}
function addTestObject() {
    let s = new Sprite("images/weight.png");
    s.x = 500;
    s.y = canvas.height - 48;
    s.visible = true;
    sprites.objects.push(s);
}
function manageTestObject() {
    sprites.objects[0].animate();
}
function drawFrame() {
    ctx.clearRect(0, 0, canvas.width, canvas.height);
    manageBackgrounds();
    manageTestObject();
    //animateMiddleGround();
     character.animate();
 }

var flags = {
    idle: 0,
    left: 1,
    right: 2,
    jump: 4,
    fall: 8
};
function makeCharacter(){
    character = new Actor();
    character.height = 64;
    character.x = Math.floor(canvas.width / 5);
    character.y = canvas.height - (character.height);
    character.blockRight = character.blockLeft = false;
    character.dropTo = null;
    let a = new Anim("images/idle10x37x64.png", 10, 37, true);
    a.pingPong = true;
    character.anims.push(a);
    a = new Anim("images/RunRight10x46x64.png", 10, 46, true);
    character.anims.push(a);
    a = new Anim("images/RunLeft10x46x64.png", 10, 46, true);
    character.anims.push(a);
    a = new Anim("images/JumpRight6x40x64.png", 6, 40, false);
    character.anims.push(a);
    a = new Anim("images/JumpLeft.png", 6, 40, false);
    character.anims.push(a);
    a = new Anim("images/LandRight.png", 4, 40, false);
    character.anims.push(a);
    a = new Anim("images/LandLeft.png", 4, 40, false);
    character.anims.push(a);
    addCharacterActions();
    character.visible = true;
    character.state = flags.idle;
 }
function addCharacterActions() {
    character.jump = function() {
        if(this.state === flags.idle || (this.state & flags.jump) || (this.state & flags.fall)) {return;}
        this.state |= flags.jump; // add jump to status
        if(this.state & flags.left){
            this.resetCurrent(4);
        } else {
            this.resetCurrent(3);
        }
        this.moveY = game.shiftDown ? game.jump * 1.3 : game.jump;
        //this.y += 6;
    };
    character.manageActions = function() {
        this.collisionDetect();
        if((this.state & flags.jump) || (this.state & flags.fall)){
            this.moveY += (this.moveY >= 14) ? 0 :game.gravity; // TV
            if(this.state & flags.jump){
                 if(this.moveY >= 0){
                     this.state ^= flags.jump;
                     this.state |= flags.fall;
                     this.moveX = (this.moveX < 0) ? -2 : 2; // slow f'wd motion
                     this.resetCurrent((this.state & flags.right) ? 5 : 6);
                 }
             } else {
                 if(this.y >= canvas.height - this.height){
                     this.moveY = 0;
                     this.y = canvas.height - this.height;
                     this.state ^= flags.fall;
                     this.resetCurrent((this.state & flags.right)? 1 : 2);
                 }
             }
            return;
        }
        if(game.rightDown && !this.blockRight){
            if(!(this.state & flags.right)){
                this.resetCurrent(1);
                this.state = flags.right;
                this.blockLeft = false;
            }
            this.moveX = 5;
        } else if(game.leftDown && ! this.blockLeft){
            if(!(this.state & flags.left)){
                this.resetCurrent(2);
                this.state = flags.left;
                this.blockRight = false;
             }
             this.moveX = -5;
        } else {
            if((this.state !== flags.idle)){
                this.resetCurrent(0);
                this.state = flags.idle;
                this.moveX = 0;
            }
        }
    };
    character.collision = function(list) {
        this.dropTo = null; //signal if nothing under character
        for(let i = 0; i < list.length; i++){
           let s = sprites.platforms[list[i].index];
           switch(list[i].type){
               case "coins":
                   // add to haul and zap
                    break;
               case "objects":
                   if(this.state === flags.right || this.state === flags.left){
                       this.block();
                   } else if(this.state & flags.fall){
                       this.state ^= flags.fall;
                       this.state |= flags.jump; // bounce
                       this.moveY = game.jump / 3;
                   }
                   break;
               case "platforms":
                   break;
               case "hazards":
                   break;
           }
       }
   };
   character.block = function() {
       this.resetCurrent(0);
       this.moveX = 0;
       if(this.state & flags.right){
           game.rightDown = false;
           this.blockRight = true;
       }
       if(this.state & flags.left){
           game.leftDown = false;
           this.blockLeft = true;
       }
       this.state = flags.idle;
   };
}
function keyPressed(e) {
    switch (game.state){
       case game.states.start:
            // manage start of game 
           break;
       case game.states.running:
           switch(e.keyCode){
                case 32:
                case 38:
                    character.jump();
                    break;
                case 37:
                    game.leftDown = true;
                    break;
                case 39:
                    game.rightDown = true;
                    break;
                case 17:
                    game.ctrlDown = true;
                    break;
                case 16:
                    game.shiftDown = true;
            }
            break;
       case game.states.end:
           // manage restart
           break;
   }
}

function keyReleased(e) {
    switch(e.keyCode) {
        case 16:
            game.shiftDown = false;
            break;
        case 17:
            game.ctrlDown = false;
            break;
        case 37:
            game.leftDown = false;
            break;
        case 39:
            game.rightDown = false;
            break;
    }
}
function initialise() {
    canvas.width = 800;
    canvas.height = 500;
    window.addEventListener("keydown", keyPressed);
    window.addEventListener("keyup", keyReleased);
    makeCharacter();
    addBackgrounds();
    addTestObject();
    //initMidBackground();
    addPlatforms();
    game.state = game.states.running; // testing
    game.tmr = setInterval(drawFrame, game.speed);
}
initialise();